from typing import Protocol, runtime_checkable
import json
from importlib import import_module
from datetime import datetime


@runtime_checkable
class Saveable(Protocol):
    """
    this protocol should be implemented in every class
    that should be save/load-able from a dict.
    keep in mind, that only native python dtypes work wiht json
    """
    @property
    def qualifiedName(self) -> tuple:
        """
        module name and class name
        """
        ...

    def toDict(self) -> dict:
        """
        converts class parameters to a python dict
        all parameters need to be in native python format
        """
        pass

    @classmethod
    def fromDict(cls: object, loadDict: dict) -> object:
        """
        loads class parameters from a dict
        """
        pass


class ModelIO(object):
    @classmethod
    def load(cls, filename: str) -> object:
        # adding the file name extension
        if not filename.endswith('.json'):
            filename += '.json'

        try: # checking if file exists
            with open(filename, 'r') as f:
                loadDict = json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(f"File {filename} not found.")

        # removing unnecessary data for object recreation
        moduleName, className = loadDict.pop('qualifiedName')  # assuming qualifiedName exists in the dict
        loadDict.pop('datetime')

        # loading the proper modules
        Module = import_module(moduleName)  # dynamically import module
        Class = getattr(Module, className)  # get class from imported module

        try: # recreate the object from dict
            return Class.fromDict(loadDict)
        except AttributeError:
            raise TypeError(f'{className} is not Saveable.')

    @classmethod
    def save(cls, object: object, filename: str) -> None:
        # checking if object is savable
        if not isinstance(object, Saveable):
            raise TypeError(f'{object.__class__.__name__} is not Saveable.')

        # converting the obejct to a dict
        saveDict = object.toDict()
        saveDict['datetime'] = datetime.now().isoformat()
        saveDict['qualifiedName'] = object.qualifiedName

        # adding the file name extension
        if not filename.endswith('.json'):
            filename += '.json'

        # saving the file
        with open(filename, 'w') as f:
            json.dump(saveDict, f, indent=4)
